import inspect
import os

from ruamel.yaml import YAML
from yefpy import yef
from yefpy.mod_data import MethodInfo, ModuleInfo, FuncInfo

yaml = YAML()


class YamlClassData:
    def __init__(self):
        self.class_name: str = ""
        self.methods_info: dict[str, MethodInfo] = {}


class YamlFuncData:
    def __init__(self, name: str, func_info: FuncInfo):
        self.name = name
        self.func_info = func_info


class YamlModData:
    def __init__(self, mod_info: ModuleInfo, classes: list[YamlClassData], functions: list[YamlFuncData]):
        self.mod_info = mod_info
        self.classes = classes
        self.functions = functions


def get_yaml_modules_data(modules_info: list[ModuleInfo]) -> list[YamlModData]:
    # yaml_modules: dict[Module, list[YamlClassData]] = {}
    mod_data: list[YamlModData] = []
    for mod_info in modules_info:
        yaml_class_data: list[YamlClassData] = []
        yaml_func_data: list[YamlFuncData] = []
        for class_info in mod_info.classes:
            yaml_class = YamlClassData()
            cls: yef.YefClass = class_info.cls
            yaml_class.class_name = cls.get_class_name()
            yaml_class.methods_info = class_info.methods_info
            yaml_class_data.append(yaml_class)
        for func_info in mod_info.functions:
            yaml_func = YamlFuncData(func_info.name, func_info)
            yaml_func_data.append(yaml_func)
        mod_data.append(YamlModData(mod_info, yaml_class_data, yaml_func_data))
        # yaml_modules[mod_info] = yaml_class_data
    # return yaml_modules
    return mod_data


class ArgTypeIsEmpty(Exception):
    def __init__(self, arg_name):
        super().__init__(f'{arg_name} argument has no type')


def _handle_init(cls: YamlClassData) -> list[dict[str, str]]:
    init_args: list[dict[str, str]] = []
    for arg in cls.methods_info["__init__"].params:
        if arg.param_type is inspect._empty and arg.name != "self":
            raise ArgTypeIsEmpty(arg.name)
        if arg.name != "self":
            init_args.append({"name": arg.name, "type": arg.param_type})
    return init_args


def _handle_class_method(cls: YamlClassData) -> dict:
    class_methods_data = {}
    for method_name, method in cls.methods_info.items():
        if method_name == "__init__":
            continue
        class_method_args = []
        class_method_type = "class"
        has_type = False
        for method_arg in method.params:
            if method_arg.name == "self":
                class_method_type = "instance"
                has_type = True
                continue
            elif not has_type:
                class_method_type = "class"
            class_method_args.append({'name': method_arg.name, 'type': method_arg.param_type})
        class_methods_data[method_name] = {'type': class_method_type, 'args': class_method_args,
                                           'output': method.return_type}
    return class_methods_data


def _handle_module_func(func: YamlFuncData) -> dict:
    args = []
    for param in func.func_info.params:
        args.append({"name": param.name, "type": param.param_type})
    return {"name": func.name, "args": args, "output": func.func_info.return_type}


def generate_yaml(ymd: list[YamlModData], filename: str, gen_dir_path: str):
    yaml_data: dict = {'modules': {}}
    for module_data in ymd:
        module = module_data.mod_info.module
        mn = module.Name
        modules = yaml_data["modules"]
        modules[mn] = {"gopkg": module.Gopkg, "filename": module.Filename,
                       "import": module.Imports, "classes": [], "functions": []}
        for cls in module_data.classes:
            class_data = {"name": cls.class_name, "args": _handle_init(cls), "methods": _handle_class_method(cls)}
            modules[mn]["classes"].append(class_data)
        for func in module_data.functions:
            modules[mn]["functions"].append(_handle_module_func(func))
        yaml_data["modules"] = modules
    if not os.path.exists(gen_dir_path):
        os.makedirs(gen_dir_path)
    with open(os.path.join(gen_dir_path, filename), 'w') as file:
        yaml.dump(yaml_data, file)



def gotype(args_info: dict, return_type: str):
    def decorator(func):
        for arg_name, arg_type in args_info.items():
            func.__annotations__[arg_name] = arg_type
        func.__annotations__['return'] = return_type
        return func
    return decorator



import importlib
import inspect
from yefpy import yef


class MethodParam:
    def __init__(self, name: str, param_type: type):
        self.name = name
        self.param_type = param_type

    def __repr__(self):
        return f"{self.name}: {self.param_type}"


class MethodInfo:
    def __init__(self, name: str, params: list[MethodParam], return_type: str):
        self.name = name
        self.params = params
        self.return_type = return_type

    def __repr__(self):
        f = ", ".join([p.__repr__() for p in self.params])
        return f"{self.name}({f})"


class ClassInfo:
    def __init__(self, cls, methods_info: dict[str, MethodInfo]):
        self.cls = cls
        self.methods_info = methods_info


class FuncInfo(MethodInfo):
    def __init__(self, name: str, params: list[MethodParam], return_type: str):
        super().__init__(name, params, return_type)


class Module:
    def __init__(self, name: str, gopkg: str, filename: str, imports: list[str]):
        self.Name = name
        self.Gopkg = gopkg
        self.Filename = filename
        self.Imports = imports


class ModuleInfo:
    def __init__(self, module: Module, classes: list[ClassInfo], functions: list[FuncInfo]):
        self.module = module
        self.classes = classes
        self.functions = functions


def get_classes(module) -> list:
    classes = []
    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj):
            classes.append(obj)
    return classes


def get_class_info(class_) -> ClassInfo:
    methods_info: dict[str, MethodInfo] = {}
    methods = [method for method in inspect.getmembers(class_) if method[0].startswith("_") == False and not method[1].
        __qualname__.startswith(yef.YefClass.__name__)]
    methods.append(("__init__", class_.__init__))
    for method in methods:
        class_methods = inspect.signature(method[1]).parameters
        method_params: list[MethodParam] = []
        for name, param in class_methods.items():
            method_params.append(MethodParam(name, param.annotation))
        methods_info[method[0]] = MethodInfo(method[0], method_params, inspect.signature(method[1]).return_annotation)
    return ClassInfo(class_, methods_info)


def get_module_functions_info(module) -> list[FuncInfo]:
    functions_info: list[FuncInfo] = []
    members = inspect.getmembers(module)
    functions = [func for _, func in members if inspect.isfunction(func) and inspect.getmodule(func) == module]
    for func in functions:
        func_params: list[MethodParam] = []
        for name, param in inspect.signature(func).parameters.items():
            func_params.append(MethodParam(name, param.annotation))
        functions_info.append(FuncInfo(func.__name__, func_params, inspect.signature(func).return_annotation))
    return functions_info


def get_modules_info(modules: list[Module]) -> list[ModuleInfo]:
    modules_info: list[ModuleInfo] = []
    for module in modules:
        module_classes_info: list[ClassInfo] = []
        imported_module = importlib.import_module(str(module.Name))
        members = inspect.getmembers(imported_module)
        local_classes = [member for name, member in members if inspect.getmodule(member) == imported_module]
        for lclass in local_classes:
            if inspect.isclass(lclass) and issubclass(lclass, yef.YefClass):
                clsInfo = get_class_info(lclass)
                module_classes_info.append(clsInfo)
        module_func_info = get_module_functions_info(imported_module)
        modules_info.append(ModuleInfo(module, module_classes_info, module_func_info))
    return modules_info



import os
from yefpy import cfg


def generate_start_sh(projects_path: list[str], gen_path: str, cgo_ldflags=cfg.CGO_LDFLAGS, cgo_cflags=cfg.CGO_CFLAGS):
    if not os.path.exists(gen_path):
        os.makedirs(gen_path)
    with open(os.path.join(gen_path, 'start.sh'), "w") as script:
        script.write(f'export CGO_LDFLAGS="{cgo_ldflags}"\n')
        script.write(f'export CGO_CFLAGS="{cgo_cflags}"\n')
        for project in projects_path:
            script.write(f'export PYTHONPATH={os.path.abspath(project)}:$PYTHONPATH\n')
